from typing import Dict, List, Tuple, Optional, Iterable, Set

from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.scheduler.heft.time_computaion import calculate_working_time
from sampo.schemas.contractor import WorkerContractorPool, Contractor
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time import Time
from sampo.schemas.types import AgentId

# stacks of time(Time) and count[int]
JustInTimeTimeline = Dict[AgentId, List[Tuple[Time, int]]]


def make_and_cache_schedule(id2swork: Dict[str, ScheduledWork],
                            workers: List[Worker],
                            contractor: Contractor,
                            inseparable_chain: List[GraphNode],
                            start_time: Time,
                            exec_times: Dict[GraphNode, Tuple[Time, Time]],
                            work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
    """
    Makes ScheduledWork object from `GraphNode` and worker list, assigned `start_end_time`
    and adds it ti given `id2swork`. Also does the same for all inseparable nodes starts from this one
    :param node:
    :param id2swork:
    :param workers:
    :param contractor:
    :param inseparable_chain:
    :param start_time:
    :param exec_times:
    :param work_estimator:
    :return:
    """
    c_ft = start_time
    for dep_node in inseparable_chain:
        # set start time as finish time of original work
        # set finish time as finish time + working time of current node with identical resources
        # (the same as in original work)
        # set the same workers on it
        # TODO Decide where this should be
        max_parent_time = max((id2swork[pnode.id].finish_time
                               for pnode in dep_node.parents),
                              default=Time(0))
        working_time = exec_times.get(dep_node, None)
        if working_time is None:
            working_time = calculate_working_time(dep_node.work_unit, workers, work_estimator)
        new_finish_time = max_parent_time + working_time

        id2swork[dep_node.id] = ScheduledWork(work_unit=dep_node.work_unit,
                                              start_end_time=(max_parent_time, new_finish_time),
                                              workers=workers,
                                              contractor=contractor)
        # change finish time for using workers
        c_ft = new_finish_time

    return c_ft


def schedule_with_time_spec(node: GraphNode,
                            id2swork: Dict[str, ScheduledWork],
                            workers: List[Worker],
                            contractor: Contractor,
                            inseparable_chain: List[GraphNode],
                            timeline: JustInTimeTimeline,
                            assigned_time: Optional[Time],
                            work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
    if assigned_time:
        st = find_min_start_time(node, workers, timeline, id2swork)
        exec_times = {n: (Time(0), assigned_time // len(inseparable_chain))
                      for n in inseparable_chain}
        return make_and_cache_schedule(id2swork, workers, contractor, inseparable_chain,
                                       st, exec_times, work_estimator)
    else:
        return schedule(node, id2swork, workers, contractor, inseparable_chain, timeline, work_estimator)


def find_min_start_time(node: GraphNode, worker_team: List[Worker],
                        timeline: JustInTimeTimeline,
                        id2swork: Dict[str, ScheduledWork]) -> Time:
    """
    Define the nearest possible start time for current job. It is equal the max value from:
    1. end time of all parent tasks
    2. time previous job off all needed workers to complete current task
    :param node: target node
    :param worker_team: worker team under testing
    :param timeline: save the end time of previous work for all workers
    :param id2swork:
    :return: found time, queue of employees without assigned workers
    """
    # if current job is the first
    if len(id2swork) == 0:
        return Time(0)
    # define the max end time of all parent tasks
    max_parent_time = max([id2swork[parent_node.id].finish_time
                           for parent_node in node.parents], default=Time(0))
    # define the max agents time when all needed workers are off from previous tasks
    max_agent_time = Time(0)

    # For each resource type
    for worker in worker_team:
        needed_count = worker.count
        offer_stack = timeline[worker.get_agent_id()]
        # Traverse list while not enough resources and grab it
        ind = len(offer_stack) - 1
        while needed_count > 0:
            offer_time, offer_count = offer_stack[ind]
            max_agent_time = max(max_agent_time, offer_time)

            if needed_count < offer_count:
                offer_count = needed_count
            needed_count -= offer_count
            ind -= 1

    return max(max_agent_time, max_parent_time)


def create_timeline(agents: WorkerContractorPool) -> JustInTimeTimeline:
    """
    Initializes basic resources timeline from agents dictionary. This structure should contain
    (sorted[descending] by free time) timeline for each type of worker defined as `AgentId`.
    Element of timeline have type `Tuple[Time, Int]`, so `Tuple[t, c]`
    means at the moment t there are c free workers of the corresponding `AgentId`
    :param agents:
    :return:
    """
    timeline: JustInTimeTimeline = {}
    for worker_type, worker_offers in agents.items():
        for worker_offer in worker_offers.values():
            timeline[worker_offer.get_agent_id()] = [(Time(0), worker_offer.count)]
    return timeline


def update_timeline(finish: Time, timeline: JustInTimeTimeline, worker_team: List[Worker]):
    """
    Adds given `worker_team` to the timeline at the moment `finish`
    :param finish:
    :param timeline:
    :param worker_team:
    :return:
    """
    # For each worker type consume the nearest available needed worker amount
    # and re-add it to the time when current work should be finished.
    # Addition performed as step in bubble-sort algorithm.
    for worker in worker_team:
        needed_count = worker.count
        worker_timeline = timeline[(worker.contractor_id, worker.name)]
        # Consume needed workers
        while needed_count > 0:
            next_time, next_count = worker_timeline.pop()
            if next_count > needed_count:
                worker_timeline.append((next_time, next_count - needed_count))
                break
            needed_count -= next_count

        # Add to the right place
        # worker_timeline.append((finish, worker.count))
        # worker_timeline.sort(reverse=True)
        worker_timeline.append((finish, worker.count))
        ind = len(worker_timeline) - 1
        while ind > 0 and worker_timeline[ind][0] > worker_timeline[ind - 1][0]:
            worker_timeline[ind], worker_timeline[ind - 1] = worker_timeline[ind - 1], worker_timeline[ind]
            ind -= 1
    return timeline


def schedule(node: GraphNode,
             id2swork: Dict[str, ScheduledWork],
             workers: List[Worker],
             contractor: Contractor,
             inseparable_chain: List[GraphNode],
             timeline: JustInTimeTimeline,
             work_estimator: Optional[WorkTimeEstimator] = None) -> Time:
    """
    Makes ScheduledWork object from `GraphNode` and worker list, computes it's `start_end_time`
    and adds it ti given `id2swork`. Also does the same for all inseparable nodes starts from this one
    :param node:
    :param id2swork:
    :param workers:
    :param contractor:
    :param inseparable_chain:
    :param timeline:
    :param work_estimator:
    :return:
    """
    st = find_min_start_time(node, workers, timeline, id2swork)
    return make_and_cache_schedule(id2swork, workers, contractor, inseparable_chain, st, {}, work_estimator)


def order_nodes_by_start_time(works: Iterable[ScheduledWork], wg: WorkGraph) -> List[str]:
    """
    Makes ScheduledWorks' ordering that satisfies:
    1. Ascending order by start time
    2. Toposort
    :param works:
    :param wg:
    :return:
    """
    res = []
    order_by_start_time = [(item.start_time, item.work_unit.id) for item in
                           sorted(works, key=lambda item: item.start_time)]

    cur_time = 0
    cur_class: Set[GraphNode] = set()
    for start_time, work in order_by_start_time:
        node = wg[work]
        if len(cur_class) == 0:
            cur_time = start_time
        if start_time == cur_time:
            cur_class.add(node)
            continue
        # TODO Perform real toposort
        cur_not_added: Set[GraphNode] = set(cur_class)
        while len(cur_not_added) > 0:
            for cur_node in cur_class:
                if any([parent_node in cur_not_added for parent_node in cur_node.parents]):
                    continue  # we add this node later
                res.append(cur_node.id)
                cur_not_added.remove(cur_node)
            cur_class = set(cur_not_added)
        cur_time = start_time
        cur_class = {node}

    cur_not_added: Set[GraphNode] = set(cur_class)
    while len(cur_not_added) > 0:
        for cur_node in cur_class:
            if any([parent_node in cur_not_added for parent_node in cur_node.parents]):
                continue  # we add this node later
            res.append(cur_node.id)
            cur_not_added.remove(cur_node)
        cur_class = set(cur_not_added)

    return res



import random
from typing import List, Dict, Iterable

import numpy as np
from deap import creator, base, tools

from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.scheduler.genetic.converter import convert_chromosome_to_schedule
from sampo.scheduler.genetic.converter import convert_schedule_to_chromosome, ChromosomeType
from sampo.scheduler.topological.base import RandomizedTopologicalScheduler
from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.graph import GraphNode, WorkGraph
from sampo.schemas.schedule import ScheduledWork
from sampo.schemas.time import Time
# create class FitnessMin, the weights = -1 means that fitness - is function for minimum
from sampo.utilities.collections import reverse_dictionary

creator.create("FitnessMin", base.Fitness, weights=(-1.0,))
creator.create("Individual", list, fitness=creator.FitnessMin)
Individual = creator.Individual


def init_toolbox(wg: WorkGraph, contractors: List[Contractor], agents: WorkerContractorPool,
                 index2node: Dict[int, GraphNode],
                 work_id2index: Dict[str, int], worker_name2index: Dict[str, int],
                 index2contractor: Dict[int, str],
                 index2contractor_obj: Dict[int, Contractor],
                 init_chromosomes: Dict[str, ChromosomeType],
                 mutate_order: float, mutate_resources: float, selection_size: int,
                 rand: random.Random,
                 spec: ScheduleSpec,
                 work_estimator: WorkTimeEstimator = None) -> base.Toolbox:
    toolbox = base.Toolbox()
    # generate initial population
    toolbox.register("generate_chromosome", generate_chromosome, wg=wg, contractors=contractors, index2node=index2node,
                     work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=reverse_dictionary(index2contractor),
                     init_chromosomes=init_chromosomes, rand=rand, work_estimator=work_estimator)

    # create from generate_chromosome function one individual
    toolbox.register("individual", tools.initRepeat, Individual, toolbox.generate_chromosome, n=1)
    # create population from individuals
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    # evaluation function
    toolbox.register("evaluate", chromosome_evaluation, index2node=index2node,
                     worker_name2index=worker_name2index,
                     index2worker_name=reverse_dictionary(worker_name2index),
                     index2contractor=index2contractor_obj, agents=agents,
                     spec=spec, work_estimator=work_estimator)
    # crossover for order
    toolbox.register("mate", mate_scheduling_order, rand=rand)
    # mutation for order. Coefficient luke one or two mutation in individual
    toolbox.register("mutate", tools.mutShuffleIndexes, indpb=mutate_order)
    # selection. Some random individuals and arranges a battle between them as a result in a continuing genus,
    # this is the best among these it
    toolbox.register("select", tools.selTournament, tournsize=selection_size)

    # mutation for resources
    toolbox.register("mutate_resources", mut_uniform_int, probability_mutate_resources=mutate_resources,
                     contractor_count=len(index2contractor), rand=rand)
    # crossover for resources
    toolbox.register("mate_resources", mate_for_resources, rand=rand)

    toolbox.register("validate", is_chromosome_correct, index2node=index2node, index2contractor=index2contractor_obj,
                     agents=agents, index2worker=reverse_dictionary(worker_name2index))
    toolbox.register("schedule_to_chromosome", convert_schedule_to_chromosome, index2node=index2node,
                     work_id2index=work_id2index, worker_name2index=worker_name2index,
                     contractor2index=reverse_dictionary(index2contractor))
    toolbox.register("chromosome_to_schedule", convert_chromosome_to_schedule, agents=agents,
                     index2node=index2node, worker_name2index=worker_name2index, index2contractor=index2contractor_obj,
                     spec=spec, work_estimator=work_estimator)
    return toolbox


def generate_chromosome(wg: WorkGraph, contractors: List[Contractor], index2node: Dict[int, GraphNode],
                  work_id2index: Dict[str, int], worker_name2index: Dict[str, int],
                  contractor2index: Dict[str, int],
                  init_chromosomes: Dict[str, ChromosomeType], rand: random.Random,
                  work_estimator: WorkTimeEstimator = None) -> ChromosomeType:
    """
    It is necessary to generate valid scheduling, which are satisfied to current dependencies
    That's why will be used the approved order of works (HEFT order and Topological sorting)
    Topological sorts are generating always different
    HEFT is always the same(now not)
    HEFT we will choose in 30% of attempts
    Topological in others
    :param contractor2index:
    :param work_estimator:
    :param contractors:
    :param wg:
    :param work_id2index:
    :param index2node:
    :param worker_name2index:
    :param rand:
    :param init_chromosomes:
    :return: chromosome
    """

    # do random choice to choose type of generation current chromosome (HEFT or Topological type)
    chance = rand.random()
    if chance < 0.2:
        chromosome = init_chromosomes["heft_end"]
    elif chance < 0.4:
        chromosome = init_chromosomes["heft_between"]
    else:
        schedule = RandomizedTopologicalScheduler(work_estimator,
                                                  int(rand.random() * 1000000)) \
            .schedule(wg, contractors)
        chromosome = convert_schedule_to_chromosome(index2node, work_id2index, worker_name2index,
                                                    contractor2index, schedule)
    return chromosome


def chromosome_evaluation(individuals: List[ChromosomeType], index2node: Dict[int, GraphNode],
                          worker_name2index: Dict[str, int], index2worker_name: Dict[int, str],
                          index2contractor: Dict[int, Contractor], agents: WorkerContractorPool,
                          spec: ScheduleSpec, work_estimator: WorkTimeEstimator = None) -> Time:
    chromosome = individuals[0]
    if is_chromosome_correct(chromosome, index2node, index2contractor, agents, index2worker_name):
        scheduled_works = convert_chromosome_to_schedule(chromosome, agents, index2node,
                                                         worker_name2index,
                                                         index2contractor,
                                                         spec, work_estimator)
        return max(scheduled_works.values(), key=ScheduledWork.finish_time_getter()).finish_time
    else:
        return Time.inf()


def is_chromosome_correct(chromosome: ChromosomeType, index2node: Dict[int, GraphNode],
                          index2contractor: Dict[int, Contractor], agents: WorkerContractorPool,
                          index2worker: Dict[int, str]) -> bool:
    return is_chromosome_order_correct(chromosome, index2node) and \
           is_chromosome_contractors_correct(chromosome, index2contractor, agents, index2worker, index2node.keys())


def is_chromosome_order_correct(chromosome: ChromosomeType, index2node: Dict[int, GraphNode]) -> bool:
    work_order = chromosome[0]
    used = set()
    for work_index in work_order:
        node = index2node[work_index]
        for parent in node.parents:
            if parent.id not in used:
                return False
        used.add(node.id)
    return True


def is_chromosome_contractors_correct(chromosome: ChromosomeType,
                                      index2contractor: Dict[int, Contractor],
                                      agents: WorkerContractorPool,
                                      index2worker: Dict[int, str],
                                      work_indices: Iterable[int]) -> bool:
    """
    Checks that assigned contractors can supply assigned workers
    :param chromosome:
    :param index2contractor:
    :param agents:
    :param index2worker:
    :param work_indices:
    :return:
    """
    for ind in work_indices:
        resources_count = chromosome[1][:-1, ind]
        resources_names = [index2worker[i] for i in range(resources_count.size)]
        contractor_ind = chromosome[1][-1, ind]
        contractor_id = index2contractor[contractor_ind].id
        for name, count in zip(resources_names, resources_count):
            if agents[name][contractor_id].count < count:
                return False
    return True


def get_order_tail(head_set: List[int], other: List[int]) -> List[int]:
    head_set = set(head_set)
    return [node for node in other if node not in head_set]


def mate_scheduling_order(ind1: List[int], ind2: List[int], rand: random.Random) -> (List[int], List[int]):
    """
    Crossover for order
    Basis crossover is cxOnePoint
    But we checked not repeated works in individual order
    :param ind1:
    :param ind2:
    :param rand:
    :return: two cross individuals
    """
    # randomly select the point where the crossover will take place
    crossover_point = rand.randint(1, len(ind1))

    ind1_new_tail = get_order_tail(ind1[:crossover_point], ind2)
    ind2_new_tail = get_order_tail(ind2[:crossover_point], ind1)

    ind1[crossover_point:] = ind1_new_tail
    ind2[crossover_point:] = ind2_new_tail

    return ind1, ind2


def mut_uniform_int(chromosome: ChromosomeType, low: np.ndarray, up: np.ndarray, type_of_worker: int,
                    probability_mutate_resources: float, contractor_count: int, rand: random.Random) -> ChromosomeType:
    """
    Mutation function for resources
    It changes selected numbers of workers in random work in certain interval for this work
    :param contractor_count:
    :param chromosome:
    :param low:
    :param up:
    :param type_of_worker:
    :param probability_mutate_resources:
    :param rand:
    :return: mutate individual
    """
    # select random number from interval from min to max from uniform distribution
    size = len(chromosome[1][type_of_worker])

    if type_of_worker == len(chromosome[1]) - 1:
        # print('Contractor mutation!')
        for i in range(size):
            if rand.random() < probability_mutate_resources:
                chromosome[1][type_of_worker][i] = rand.randint(0, contractor_count - 1)
        return chromosome

    # change in this interval in random number from interval
    for i, xl, xu in zip(range(size), low, up):
        if rand.random() < probability_mutate_resources:
            chromosome[1][type_of_worker][i] = rand.randint(xl, xu)

    return chromosome


def mate_for_resources(ind1: np.ndarray, ind2: np.ndarray, rand: random.Random) -> (
        np.ndarray, np.ndarray):
    """
    CxOnePoint for resources
    :param ind1:
    :param ind2:
    :param rand:
    :return:
    """
    cxpoint = rand.randint(1, len(ind1))
    ind1[cxpoint:], ind2[cxpoint:] = ind2[cxpoint:], ind1[cxpoint:]
    return ind1, ind2



from operator import attrgetter
from typing import List, Dict, Set, Iterable

from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.scheduler.utils.just_in_time_timeline import create_timeline, find_min_start_time, update_timeline
from sampo.schemas.contractor import WorkerContractorPool
from sampo.schemas.graph import GraphNode
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.schedule import ScheduledWork
from sampo.utilities.collections import build_index

PRIORITY_SHUFFLE_RADIUS = 0.5


def get_swap_candidates(node: GraphNode,
                        node_index: int,
                        candidates: Iterable[GraphNode],
                        node2ind: Dict[GraphNode, int],
                        processed: Set[GraphNode]):
    """
    Abstract function to find nodes that can be swapped
    with given node without breaking topological order
    :param node: target node
    :param node_index: index of target node in global sequence
    :param candidates: list of candidates to swapping
    :param node2ind: a dict from node to it's index
    :param processed: a set of nodes that should not be swapped yet
    """
    cur_children: Set[GraphNode] = node.children_set

    def is_candidate_accepted(candidate: GraphNode) -> bool:
        if candidate in cur_children or candidate in processed:
            return False
        candidate_ind = node2ind[candidate]
        for child in cur_children:
            if node2ind.get(child, 0) >= candidate_ind:  # we have a child between us and candidate
                return False
        candidate_parents = candidate.parents
        for parent in candidate_parents:
            if node2ind.get(parent, 0) <= node_index:  # candidate has a parent between us and candidate
                return False
        return True

    return [candidate for candidate in candidates if is_candidate_accepted(candidate)]


def shuffle_local_sequence(seq: List[GraphNode],
                           start_index: int,
                           end_index: int):
    """
    Params: sub-seq for optimizing, work time estimator
    This performs just small shuffle that not break topological order
    :param seq:
    :param start_index:
    :param end_index:
    :return:
    """
    # TODO Examine what is better: perform shuffling in nearly placed sub-seq or in whole sequence
    # node2cost = {node: work_priority(node, calculate_working_time_cascade, work_estimator) for node in sub_seq}

    # preprocessing
    node2ind: Dict[GraphNode, int] = {node: start_index + ind for ind, node in enumerate(seq[start_index:end_index])}

    # temporary for usability measurement
    swapped = 0

    processed: Set[GraphNode] = set()
    for i in reversed(range(start_index, end_index)):
        node = seq[i]
        if node in processed:
            continue
        # cur_cost = node2cost[node]
        chain_candidates: List[GraphNode] = seq[start_index:i]

        accepted_candidates = get_swap_candidates(node, i, chain_candidates, node2ind, processed)

        if accepted_candidates:
            chain_candidate = accepted_candidates[0]
            swap_idx = node2ind[chain_candidate]
            seq[i], seq[swap_idx] = seq[swap_idx], seq[i]
            # print(f'Swapped {i} and {swap_idx}')
            processed.add(chain_candidate)
            node2ind[chain_candidate] = i
            node2ind[node] = swap_idx
            swapped += 1

        processed.add(node)
    print(f'Swapped {swapped} times!')


def parallelize_local_sequence(seq: List[GraphNode],
                               start_index: int,
                               end_index: int,
                               id2schedule: Dict[str, ScheduledWork]):
    """
    This method finds near placed works and turns it to run in parallel.
    It will take effect only if it's launched after scheduling
    :param seq: nodes to optimize
    :param start_index: the start of target sub-seq
    :param end_index: the end of target sub-seq
    :param id2schedule: index of scheduled works
    :return:
    """
    # preprocessing
    node2ind: Dict[GraphNode, int] = {node: start_index + ind for ind, node in enumerate(seq[start_index:end_index])}

    processed: Set[GraphNode] = set()

    for i in reversed(range(start_index, end_index)):
        node = seq[i]
        if node in processed:
            continue
        chain_candidates: List[GraphNode] = seq[0:i]
        accepted_candidates = get_swap_candidates(node, i, chain_candidates, node2ind, processed)

        my_schedule: ScheduledWork = id2schedule[node.id]
        my_workers: Dict[str, Worker] = build_index(my_schedule.workers, attrgetter('name'))
        my_schedule_reqs: Dict[str, WorkerReq] = build_index(my_schedule.work_unit.worker_reqs, attrgetter('kind'))

        new_my_workers = {}

        # now accepted_candidates is a list of nodes that can(according to dependencies) run in parallel
        for candidate in accepted_candidates:
            candidate_schedule = id2schedule[candidate.id]

            candidate_schedule_reqs: Dict[str, WorkerReq] = build_index(candidate_schedule.work_unit.worker_reqs,
                                                                        attrgetter('kind'))

            new_candidate_workers: Dict[str, int] = {}

            satisfy = True

            for candidate_worker in candidate_schedule.workers:
                my_worker = my_workers.get(candidate_worker.name, None)
                if my_worker is None:  # these two works are not compete for this worker
                    continue

                need_me = my_workers[candidate_worker.name].count
                need_candidate = candidate_worker.count

                total = need_me + need_candidate
                my_req = my_schedule_reqs[candidate_worker.name]
                candidate_req = candidate_schedule_reqs[candidate_worker.name]
                needed_min = my_req.min_count + candidate_req.min_count

                if needed_min > total:  # these two works can't run in parallel
                    satisfy = False
                    break

                candidate_worker_count = candidate_req.min_count
                my_worker_count = my_req.min_count
                total -= needed_min

                add_me = min(my_req.max_count, total // 2)
                add_candidate = min(candidate_req.max_count, total - add_me)

                my_worker_count += add_me
                candidate_worker_count += add_candidate

                new_my_workers[candidate_worker.name] = my_worker_count
                new_candidate_workers[candidate_worker.name] = candidate_worker_count

            if satisfy:  # replacement found, apply changes and leave candidates bruteforce
                print(f'Found! {candidate.work_unit.name} {node.work_unit.name}')
                for worker in my_schedule.workers:
                    worker_count = new_my_workers.get(worker.name, None)
                    if worker_count is not None:
                        worker.count = worker_count
                for worker in candidate_schedule.workers:
                    worker_count = new_candidate_workers.get(worker.name, None)
                    if worker_count is not None:
                        worker.count = worker_count
                # candidate_schedule.start_time = my_schedule.start_time
                break


def recalc_schedule(seq: Iterable[GraphNode],
                    id2scheduled_work: Dict[str, ScheduledWork],
                    agents: WorkerContractorPool,
                    work_estimator: WorkTimeEstimator = None):
    """
    Recalculates duration and start-finish times in the whole given `seq`.
    This will be useful to call after `parallelize_local_sequence` method
    or other methods that can change the appointed set of workers.
    :param seq: scheduled works to process
    :param id2scheduled_work:
    :param agents:
    :param work_estimator: an optional WorkTimeEstimator object to estimate time of work
    """

    timeline = create_timeline(agents)
    id2scheduled_work_new: Dict[str, ScheduledWork] = {}

    for node in seq:
        node_schedule = id2scheduled_work[node.id]
        st = find_min_start_time(node, node_schedule.workers, timeline, id2scheduled_work_new)
        ft = st + node_schedule.get_actual_duration(work_estimator)
        update_timeline(ft, timeline, node_schedule.workers)
        node_schedule.start_end_time = (st, ft)
        id2scheduled_work_new[node.id] = node_schedule


def optimize_local_sequence(seq: List[GraphNode],
                            start_ind: int,
                            end_ind: int,
                            work_estimator: WorkTimeEstimator = None):
    # TODO Try to find sets of works with nearly same resources and turn in to run in parallel or vice-versa
    pass



from collections import namedtuple
from math import inf
from typing import Dict, List, Set


def is_sequence_correct(adj_matrix: Dict[str, Dict[str, List[str]]], sequence: List[str]) -> (
        bool, Dict[str, Set[str]]):
    """
    Checks the sequence for correctness and builds an adjacency matrix containing
    all hard links and soft links performed
    :param adj_matrix: adjacency matrix, where for each key the value is a dictionary {'hard': [], 'soft': []}
    :param sequence: vertex sequence from the chromosome
    :return:
        is_correct (bool): True if the sequence is correct, otherwise False
        correct_adj_matrix (Dict[str, Set[str]]): if is_correct is True, then the matrix is from the description, otherwise None
    """
    # Check that all ids are present in the sequence and exactly once
    if len(set(adj_matrix.keys())) != len(set(sequence)):
        return False, None

    # searching for uncompleted hard ties
    used: Set[str] = set()
    correct_adj_matrix: Dict[str, Set[str]] = dict()
    for s in sequence:
        used.add(s)
        hard_children = set(adj_matrix[s]['hard'])
        if len(used & hard_children) > 0:
            return False, None
        correct_adj_matrix[s] = hard_children
        soft_children = set(adj_matrix[s]['soft']) - used
        correct_adj_matrix[s] |= soft_children
    return True, correct_adj_matrix


def soft_metric_score(adj_matrix: Dict[str, Dict[str, List[str]]], sequence: List[str]) -> float:
    """
        counts the metric of the number of soft ties
        :param adj_matrix: adjacency matrix, where for each key the value is a dictionary {'hard': [], 'soft': []}
        :param sequence: vertex sequence from the chromosome
        :return:
            score (float): The number of executed soft links if the sequence is correct, otherwise math.inf
    """
    # Check that all ids are present in the sequence and exactly once
    is_correct, correct_adj = is_sequence_correct(adj_matrix, sequence)
    if not is_correct:
        return inf

    score = sum(len(correct_adj[key] - set(adj_matrix[key]['hard'])) for key in sequence)
    return score


def mean_path_score(adj_matrix: Dict[str, Dict[str, List[str]]], sequence: List[str]) -> float:
    """
        counts the metric of the mean length path
        :param adj_matrix: adjacency matrix, where for each key the value is a dictionary {'hard': [], 'soft': []}
        :param sequence: vertex sequence from the chromosome
        :return:
            score (float): The mean length path in the graph if the sequence is correct, otherwise math.inf
    """
    # Check that all ids are present in the sequence and exactly once
    is_correct, correct_adj = is_sequence_correct(adj_matrix, sequence)
    if not is_correct:
        return inf

    PathCounter = namedtuple('PathCounter', 'sum count')
    counter: Dict[str, PathCounter] = dict()

    def dfs(v_id: str):
        if v_id in counter:
            return
        if len(correct_adj[v_id]) == 0:
            counter[v_id] = PathCounter(0, 1)
            return
        path_sum, path_count = 0, 0
        for child in correct_adj[v_id]:
            dfs(child)
            child_path = counter[child]
            path_sum += child_path.sum + child_path.count
            path_count += child_path.count
        counter[v_id] = PathCounter(path_sum, path_count)
        return

    res_sum, res_count = 0, 0
    for s_id in sequence:
        if s_id not in counter:
            dfs(s_id)
            res_sum += counter[s_id].sum
            res_count += counter[s_id].count
    return res_sum / res_count


if __name__ == "__main__":
    test_adj_matrix = {
        '1': {'hard': ['2', '3', '4'], 'soft': ['5', ]},
        '2': {'hard': ['5', ], 'soft': []},
        '3': {'hard': ['11'], 'soft': []},
        '4': {'hard': ['5'], 'soft': ['3']},
        '5': {'hard': [], 'soft': []},
        '6': {'hard': ['7', '8'], 'soft': ['9']},
        '7': {'hard': ['10'], 'soft': []},
        '8': {'hard': ['9'], 'soft': []},
        '9': {'hard': ['10'], 'soft': ['7']},
        '10': {'hard': [], 'soft': []},
        '11': {'hard': ['5'], 'soft': ['2']},
        '12': {'hard': ['15'], 'soft': ['14']},
        '13': {'hard': ['15'], 'soft': ['14']},
        '14': {'hard': ['15'], 'soft': []},
        '15': {'hard': [], 'soft': []},
    }

    slice_adj = {key: test_adj_matrix[key] for key in ['12', '13', '14', '15']}
    short_tests = [
        (slice_adj, ['12', '13', '14', '15'],
         (True, {'12': {'14', '15'}, '13': {'14', '15'}, '14': {'15'}, '15': set()}), 2, 1.5),
        (slice_adj, ['12', '13', '14', '14'], (False, None), inf, inf),
        (slice_adj, ['12', '13', '15', '14'], (False, None), inf, inf),
        (slice_adj, ['14', '13', '12', '15'],
         (True, {'14': {'15'}, '13': {'15'}, '12': {'15'}, '15': set()}), 0, 1.0),
    ]
    print(any(is_sequence_correct(test[0], test[1]) == test[2] for test in short_tests))
    print(any(soft_metric_score(test[0], test[1]) == test[3] for test in short_tests))
    print(any(mean_path_score(test[0], test[1]) == test[4] for test in short_tests))



from abc import ABC
from bisect import bisect_left, bisect_right

from sortedcontainers import SortedKeyList

# TODO: describe the class (description)
class ExtendedSortedList(SortedKeyList, ABC):
    # TODO: describe the function (parameters (old), return type)
    def merge(self, old, value, merger):
        """
        Merges `value` and `old` and places the result to the place of `old` in sorted list.
        This should NOT modify key of object.

        Runtime complexity: `O(log(n))` -- approximate.

        :param merger: a function that takes old and new value and merges it into new one
        :param value: value to merge

        """
        _lists = self._lists
        _keys = self._keys
        _maxes = self._maxes

        key = self._key(old)

        increased = True

        if _maxes:
            pos = bisect_left(_maxes, key)

            # def insert_or_merge():
            #     idx = bisect_right(_keys[pos], key)
            #     old_key = _keys[pos][idx]
            #     if old_key == key:
            #         _lists[pos][idx] = merger(_lists[pos][idx], value)
            #     else:
            #         _lists[pos].insert(idx, value)
            #         _keys[pos].insert(idx, key)

            if pos == len(_maxes):
                pos -= 1

                old_key = _keys[pos][-1]
                if old_key == key:
                    new_value = merger(_lists[pos][-1], value)
                    _lists[pos][-1] = new_value
                    # _keys[pos][-1] = self._key(new_value)
                    increased = False
                else:
                    _lists[pos].append(value)
                    _keys[pos].append(key)
                _maxes[pos] = key
            else:
                idx = bisect_right(_keys[pos], key)
                old_key = _keys[pos][idx - 1]
                if idx != 0 and old_key == key:
                    new_value = merger(_lists[pos][idx - 1], value)
                    _lists[pos][idx - 1] = new_value
                    # _keys[pos][idx - 1] = self._key(new_value)
                    increased = False
                else:
                    _lists[pos].insert(idx, value)
                    _keys[pos].insert(idx, key)

            if increased:
                self._expand(pos)
        else:
            _lists.append([value])
            _keys.append([key])
            _maxes.append(key)

        if increased:
            self._len += 1



from typing import List, Optional, Dict, Iterable

from sampo.metrics.resources_in_time.base import ResourceOptimizer
from sampo.scheduler.base import SchedulerType
from sampo.scheduler.heft.base import HEFTScheduler
from sampo.scheduler.heft.prioritization import prioritization
from sampo.scheduler.heft.time_computaion import calculate_working_time_cascade
from sampo.scheduler.resource.coordinate_descent import CoordinateDescentResourceOptimizer
from sampo.scheduler.utils.momentum_timeline import create_timeline, find_min_start_time, schedule_with_time_spec
from sampo.scheduler.utils.multi_contractor import get_best_contractor_and_worker_borders
from sampo.schemas.contractor import Contractor, get_worker_contractor_pool
from sampo.schemas.graph import WorkGraph, GraphNode
from sampo.schemas.schedule import Schedule
from sampo.schemas.schedule_spec import ScheduleSpec
from sampo.schemas.scheduled_work import ScheduledWork
from sampo.schemas.time_estimator import WorkTimeEstimator
from sampo.utilities.base_opt import dichotomy_int
from sampo.utilities.validation import validate_schedule


class HEFTBetweenScheduler(HEFTScheduler):

    def __init__(self,
                 scheduler_type: SchedulerType = SchedulerType.HEFTAddBetween,
                 resource_optimizer: ResourceOptimizer = CoordinateDescentResourceOptimizer(dichotomy_int),
                 work_estimator: Optional[WorkTimeEstimator or None] = None):
        super().__init__(scheduler_type, resource_optimizer, work_estimator)

    def schedule(self, wg: WorkGraph,
                 contractors: List[Contractor],
                 spec: ScheduleSpec = ScheduleSpec(),
                 validate: bool = False) \
            -> Schedule:
        ordered_nodes = prioritization(wg, self.work_estimator)

        schedule = Schedule.from_scheduled_works(
            self.build_scheduler(ordered_nodes, contractors, spec, self.work_estimator),
            wg
        )

        if validate:
            validate_schedule(schedule, wg, contractors)

        return schedule

    def build_scheduler(self, ordered_nodes: List[GraphNode],
                        contractors: List[Contractor],
                        spec: ScheduleSpec,
                        work_estimator: WorkTimeEstimator = None) \
            -> Iterable[ScheduledWork]:
        """
        Find optimal number of workers who ensure the nearest finish time.
        Finish time is combination of two dependencies: max finish time, max time of waiting of needed workers
        This is selected by iteration from minimum possible numbers of workers until then the finish time is decreasing
        :param ordered_nodes:
        :param contractors:
        :param spec: spec for current scheduling
        :param work_estimator:
        :return:
        """
        worker_pool = get_worker_contractor_pool(contractors)
        # dict for writing parameters of completed_jobs
        node2swork: Dict[GraphNode, ScheduledWork] = {}
        # list for support the queue of workers
        timeline = create_timeline(ordered_nodes, contractors)
        # add to queue all available workers

        for index, node in enumerate(reversed(ordered_nodes)):  # the tasks with the highest rank will be done first
            work_unit = node.work_unit
            work_spec = spec.get_work_spec(work_unit.id)
            if node in node2swork:  # here
                continue

            inseparable_chain = node.get_inseparable_chain() if node.get_inseparable_chain() is not None else [node]

            min_count_worker_team, max_count_worker_team, contractor, workers \
                = get_best_contractor_and_worker_borders(worker_pool, contractors, work_unit.worker_reqs)

            best_worker_team = [worker.copy() for worker in workers]

            def get_finish_time(worker_team):
                return find_min_start_time(timeline[contractor.id], node, node2swork, inseparable_chain,
                                           best_worker_team, work_estimator)[0] \
                       + calculate_working_time_cascade(node, worker_team, work_estimator)

            # apply worker team spec
            self.optimize_resources_using_spec(work_unit, best_worker_team, work_spec,
                                               lambda optimize_array: self.resource_optimizer.optimize_resources(
                                                   worker_pool, best_worker_team,
                                                   optimize_array,
                                                   min_count_worker_team, max_count_worker_team,
                                                   get_finish_time))

            # finish scheduling with time spec
            schedule_with_time_spec(index, node, node2swork, inseparable_chain, timeline, best_worker_team, contractor,
                                    work_spec.assigned_time, work_estimator)

        # parallelize_local_sequence(ordered_nodes, 0, len(ordered_nodes), node2swork)
        # recalc_schedule(reversed(ordered_nodes), node2swork, agents, work_estimator)

        return node2swork.values()



from typing import List

import numpy as np

from sampo.schemas.contractor import Contractor, WorkerContractorPool
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker


def get_best_contractor_and_worker_borders(agents: WorkerContractorPool,
                                           contractors: List[Contractor],
                                           work_reqs: List[WorkerReq]) \
        -> (np.ndarray, np.ndarray, Contractor, List[Worker]):
    """
    Define for each job each type of workers the min and max possible number of workers
    For max number of workers max is define as minimum from max possible numbers at all and max possible for current job
    :param agents: from all project
    :param contractors:
    :param work_reqs:
    :param timeline:
    :return:
    """
    n = len(work_reqs)
    min_worker_team = np.zeros(n, dtype=int)
    max_worker_team = np.zeros(n, dtype=int)
    workers: List[Worker] = []
    contractor: Contractor = contractors[0]
    min_sum = 1e50

    # TODO Entry point for multi-contractor
    # TODO When starting implementation of multi-contractor,
    #  see this algo about searching the most relevant contractor
    for c in contractors:
        cur_min_worker_team = np.zeros(n, dtype=int)
        cur_max_worker_team = np.zeros(n, dtype=int)
        cur_workers = []
        satisfied = True

        for i, req in enumerate(work_reqs):
            offers = agents[req.kind]

            cur_offers = list(filter(lambda x: x.contractor_id == c.id, offers.values()))
            if len(cur_offers) == 0 or cur_offers[0].count < req.min_count:
                satisfied = False
                break

            cur_min_worker_team[i] = req.min_count
            cur_max_worker_team[i] = min(req.max_count, cur_offers[0].count)
            cur_workers.append(cur_offers[0])

        cur_sum = np.sum(cur_max_worker_team - cur_min_worker_team)
        if satisfied and cur_sum < min_sum:
            min_worker_team = cur_min_worker_team
            max_worker_team = cur_max_worker_team
            workers = cur_workers
            min_sum = cur_sum
            contractor = c

    return min_worker_team, max_worker_team, contractor, workers



import queue
import uuid
from random import Random
from typing import Optional
from uuid import uuid4

from sampo.schemas.graph import GraphNode

# TODO: describe the function (description, parameters, return type)
def uuid_str(rand: Optional[Random] = None) -> str:
    ans = uuid4() if rand is None else uuid.UUID(int=rand.getrandbits(128))
    return str(ans)

# TODO: describe the function (description, parameters, return type)
def count_node_ancestors(finish: GraphNode, root: GraphNode) -> int:
    q = queue.Queue()
    count = 0
    used = set()
    used.add(root)
    q.put(finish)
    while not q.empty():
        node = q.get()
        for parent in node.parents:
            if parent in used:
                continue
            used.add(parent)
            q.put(parent)
            count += 1

    return count

# TODO: describe the function (description, parameters, return type)
def binary_search(a, x, lo=0, hi=None):
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) >> 1
        if a[mid] < x:
            lo = mid + 1
        else:
            hi = mid
    return lo

# TODO: describe the function (description, parameters, return type)
def binary_search_reversed(a, x, lo=0, hi=None):
    if hi is None:
        hi = len(a)
    while lo < hi:
        mid = (lo + hi) >> 1
        if a[mid] > x:
            lo = mid + 1
        else:
            hi = mid
    return lo



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['sampo',
 'sampo.generator',
 'sampo.generator.config',
 'sampo.generator.enviroment',
 'sampo.generator.pipeline',
 'sampo.metrics',
 'sampo.metrics.resources_in_time',
 'sampo.scheduler',
 'sampo.scheduler.genetic',
 'sampo.scheduler.heft',
 'sampo.scheduler.heft_between',
 'sampo.scheduler.resource',
 'sampo.scheduler.topological',
 'sampo.scheduler.utils',
 'sampo.schemas',
 'sampo.structurator',
 'sampo.utilities',
 'sampo.utilities.generation',
 'sampo.utilities.sampler',
 'sampo.utilities.visualization']

package_data = \
{'': ['*']}

install_requires = \
['blis>=0.7.7,<0.8.0',
 'catalogue>=2.0.7,<2.1.0',
 'catboost>=1.0.5,<1.1.0',
 'certifi>=2021.10.8,<2021.11.0',
 'charset-normalizer>=2.0.12,<2.1.0',
 'click>=8.1.3,<8.2.0',
 'cycler>=0.11.0,<0.12.0',
 'cymem>=2.0.6,<2.1.0',
 'dash>=2.4.1,<2.5.0',
 'dataclasses>=0.6,<1.0',
 'deap>=1.3.1,<1.4.0',
 'flask>=2.1.2,<2.2.0',
 'fonttools>=4.33.3,<4.34.0',
 'graphviz>=0.20,<1.0',
 'idna>=3.3,<4.0',
 'importlib-metadata>=4.11.3,<4.12.0',
 'ipython>=8.6.0,<9.0.0',
 'itsdangerous>=2.1.2,<2.2.0',
 'jinja2>=3.1.2,<3.2.0',
 'joblib>=1.1.0,<1.2.0',
 'kiwisolver>=1.4.2,<1.5.0',
 'langcodes>=3.3.0,<3.4.0',
 'markupsafe>=2.1.1,<2.2.0',
 'matplotlib>=3.5.1,<3.6.0',
 'murmurhash>=1.0.7,<1.1.0',
 'numpy-indexed>=0.3.5,<0.4.0',
 'numpy>=1.22.3,<1.23.0',
 'packaging>=21.3,<22.0',
 'pandas>=1.4.1,<1.5.0',
 'pathy>=0.6.1,<0.7.0',
 'pillow>=9.1.0,<9.2.0',
 'plotly>=5.8.0,<5.9.0',
 'preshed>=3.0.6,<3.1.0',
 'pydantic>=1.8.2,<1.9.0',
 'pyparsing>=3.0.8,<3.1.0',
 'pytest>=7.1.2,<7.2.0',
 'python-dateutil>=2.8.2,<2.9.0',
 'pytz>=2022.1,<2023.0',
 'reportlab>=3.5.67,<3.6.0',
 'requests>=2.27.1,<2.28.0',
 'scikit-learn>=1.0.2,<1.1.0',
 'scipy>=1.8.1,<1.9.0',
 'seaborn>=0.11.2,<0.12.0',
 'setuptools>=58.0.4,<58.1.0',
 'six>=1.16.0,<1.17.0',
 'smart-open>=5.2.1,<5.3.0',
 'sortedcontainers>=2.4.0,<2.5.0',
 'spacy-legacy>=3.0.9,<3.1.0',
 'spacy-loggers>=1.0.2,<1.1.0',
 'srsly>=2.4.3,<2.5.0',
 'tenacity>=8.0.1,<8.1.0',
 'thinc>=8.0.15,<8.1.0',
 'threadpoolctl>=3.1.0,<3.2.0',
 'toposort>=1.7,<2.0',
 'tqdm>=4.64.0,<4.65.0',
 'typer>=0.4.1,<0.5.0',
 'typing-extensions>=4.2.0,<4.3.0',
 'urllib3>=1.26.9,<1.27.0',
 'wasabi>=0.9.1,<0.10.0',
 'werkzeug>=2.1.2,<2.2.0',
 'zipp>=3.8.0,<3.9.0']

setup_kwargs = {
    'name': 'sampo',
    'version': '0.1.0.27',
    'description': '',
    'long_description': '# SAMPO — Scheduler for Adaptive Manufacturing Processes Optimization\n\n## Описание планировщика\n\nПланировщик для адаптивной оптимизации производственных процессов включает в себя набор алгоритмов интеллектуального анализа и построения расписаний задач производственных процессов с учетом ресурсных и прочих ограничений, накладываемых предметной областью.\n\nОн позволяет эффективно планировать производственные задачи и назначать ресурсы, оптимизируя результат планирования по требуемым метрикам.\n',
    'author': 'iAirLab',
    'author_email': 'iairlab@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)
